Intel-XE Archive on lore.kernel.org
 help / color / mirror / Atom feed
From: Matthew Brost <matthew.brost@intel.com>
To: intel-xe@lists.freedesktop.org, dri-devel@lists.freedesktop.org
Cc: himal.prasad.ghimiray@intel.com, apopple@nvidia.com,
	airlied@gmail.com, thomas.hellstrom@linux.intel.com,
	simona.vetter@ffwll.ch, felix.kuehling@amd.com, dakr@kernel.org
Subject: [PATCH v7 10/32] drm/xe: Add dma_addr res cursor
Date: Wed,  5 Mar 2025 17:26:35 -0800	[thread overview]
Message-ID: <20250306012657.3505757-11-matthew.brost@intel.com> (raw)
In-Reply-To: <20250306012657.3505757-1-matthew.brost@intel.com>

From: Thomas Hellström <thomas.hellstrom@linux.intel.com>

Add dma_addr res cursor which walks an array of drm_pagemap_dma_addr.
Useful for SVM ranges and programing page tables.

v3:
 - Better commit message (Thomas)
 - Use new drm_pagemap.h location
v7:
 - Fix kernel doc (CI)

Signed-off-by: Matthew Brost <matthew.brost@intel.com>
Signed-off-by: Thomas Hellström <thomas.hellstrom@linux.intel.com>
Reviewed-by: Matthew Brost <matthew.brost@intel.com>
---
 drivers/gpu/drm/xe/xe_res_cursor.h | 123 ++++++++++++++++++++++++++++-
 drivers/gpu/drm/xe/xe_svm.h        |   4 +
 2 files changed, 125 insertions(+), 2 deletions(-)

diff --git a/drivers/gpu/drm/xe/xe_svm.h b/drivers/gpu/drm/xe/xe_svm.h
index ac79b208304d..d361a78a6839 100644
--- a/drivers/gpu/drm/xe/xe_svm.h
+++ b/drivers/gpu/drm/xe/xe_svm.h
@@ -6,6 +6,10 @@
 #ifndef _XE_SVM_H_
 #define _XE_SVM_H_
 
+#include <drm/drm_pagemap.h>
+
+#define XE_INTERCONNECT_VRAM DRM_INTERCONNECT_DRIVER
+
 struct xe_vm;
 
 #if IS_ENABLED(CONFIG_DRM_GPUSVM)
diff --git a/drivers/gpu/drm/xe/xe_res_cursor.h b/drivers/gpu/drm/xe/xe_res_cursor.h
index dca374b6521c..d1a403cfb628 100644
--- a/drivers/gpu/drm/xe/xe_res_cursor.h
+++ b/drivers/gpu/drm/xe/xe_res_cursor.h
@@ -26,6 +26,7 @@
 
 #include <linux/scatterlist.h>
 
+#include <drm/drm_pagemap.h>
 #include <drm/ttm/ttm_placement.h>
 #include <drm/ttm/ttm_range_manager.h>
 #include <drm/ttm/ttm_resource.h>
@@ -34,17 +35,38 @@
 #include "xe_bo.h"
 #include "xe_device.h"
 #include "xe_macros.h"
+#include "xe_svm.h"
 #include "xe_ttm_vram_mgr.h"
 
 /* state back for walking over vram_mgr, stolen_mgr, and gtt_mgr allocations */
 struct xe_res_cursor {
+	/** @start: Start of cursor */
 	u64 start;
+	/** @size: Size of the current segment. */
 	u64 size;
+	/** @remaining: Remaining bytes in cursor */
 	u64 remaining;
+	/** @node: Opaque point current node cursor */
 	void *node;
+	/** @mem_type: Memory type */
 	u32 mem_type;
+	/** @sgl: Scatterlist for cursor */
 	struct scatterlist *sgl;
+	/** @dma_addr: Current element in a struct drm_pagemap_device_addr array */
+	const struct drm_pagemap_device_addr *dma_addr;
+	/** @mm: Buddy allocator for VRAM cursor */
 	struct drm_buddy *mm;
+	/**
+	 * @dma_start: DMA start address for the current segment.
+	 * This may be different to @dma_addr.addr since elements in
+	 * the array may be coalesced to a single segment.
+	 */
+	u64 dma_start;
+	/** @dma_seg_size: Size of the current DMA segment. */
+	u64 dma_seg_size;
 };
 
 static struct drm_buddy *xe_res_get_buddy(struct ttm_resource *res)
@@ -70,6 +92,7 @@ static inline void xe_res_first(struct ttm_resource *res,
 				struct xe_res_cursor *cur)
 {
 	cur->sgl = NULL;
+	cur->dma_addr = NULL;
 	if (!res)
 		goto fallback;
 
@@ -141,6 +164,36 @@ static inline void __xe_res_sg_next(struct xe_res_cursor *cur)
 	cur->sgl = sgl;
 }
 
+/**
+ * __xe_res_dma_next() - Advance the cursor when end-of-segment is reached
+ * @cur: The cursor
+ */
+static inline void __xe_res_dma_next(struct xe_res_cursor *cur)
+{
+	const struct drm_pagemap_device_addr *addr = cur->dma_addr;
+	u64 start = cur->start;
+
+	while (start >= cur->dma_seg_size) {
+		start -= cur->dma_seg_size;
+		addr++;
+		cur->dma_seg_size = PAGE_SIZE << addr->order;
+	}
+	cur->dma_start = addr->addr;
+
+	/* Coalesce array_elements */
+	while (cur->dma_seg_size - start < cur->remaining) {
+		if (cur->dma_start + cur->dma_seg_size != addr[1].addr ||
+		    addr->proto != addr[1].proto)
+			break;
+		addr++;
+		cur->dma_seg_size += PAGE_SIZE << addr->order;
+	}
+
+	cur->dma_addr = addr;
+	cur->start = start;
+	cur->size = cur->dma_seg_size - start;
+}
+
 /**
  * xe_res_first_sg - initialize a xe_res_cursor with a scatter gather table
  *
@@ -160,11 +213,42 @@ static inline void xe_res_first_sg(const struct sg_table *sg,
 	cur->start = start;
 	cur->remaining = size;
 	cur->size = 0;
+	cur->dma_addr = NULL;
 	cur->sgl = sg->sgl;
 	cur->mem_type = XE_PL_TT;
 	__xe_res_sg_next(cur);
 }
 
+/**
+ * xe_res_first_dma - initialize a xe_res_cursor with dma_addr array
+ *
+ * @dma_addr: struct drm_pagemap_device_addr array to walk
+ * @start: Start of the range
+ * @size: Size of the range
+ * @cur: cursor object to initialize
+ *
+ * Start walking over the range of allocations between @start and @size.
+ */
+static inline void xe_res_first_dma(const struct drm_pagemap_device_addr *dma_addr,
+				    u64 start, u64 size,
+				    struct xe_res_cursor *cur)
+{
+	XE_WARN_ON(!dma_addr);
+	XE_WARN_ON(!IS_ALIGNED(start, PAGE_SIZE) ||
+		   !IS_ALIGNED(size, PAGE_SIZE));
+
+	cur->node = NULL;
+	cur->start = start;
+	cur->remaining = size;
+	cur->dma_seg_size = PAGE_SIZE << dma_addr->order;
+	cur->dma_start = 0;
+	cur->size = 0;
+	cur->dma_addr = dma_addr;
+	__xe_res_dma_next(cur);
+	cur->sgl = NULL;
+	cur->mem_type = XE_PL_TT;
+}
+
 /**
  * xe_res_next - advance the cursor
  *
@@ -191,6 +275,12 @@ static inline void xe_res_next(struct xe_res_cursor *cur, u64 size)
 		return;
 	}
 
+	if (cur->dma_addr) {
+		cur->start += size;
+		__xe_res_dma_next(cur);
+		return;
+	}
+
 	if (cur->sgl) {
 		cur->start += size;
 		__xe_res_sg_next(cur);
@@ -232,6 +322,35 @@ static inline void xe_res_next(struct xe_res_cursor *cur, u64 size)
  */
 static inline u64 xe_res_dma(const struct xe_res_cursor *cur)
 {
-	return cur->sgl ? sg_dma_address(cur->sgl) + cur->start : cur->start;
+	if (cur->dma_addr)
+		return cur->dma_start + cur->start;
+	else if (cur->sgl)
+		return sg_dma_address(cur->sgl) + cur->start;
+	else
+		return cur->start;
+}
+
+/**
+ * xe_res_is_vram() - Whether the cursor current dma address points to
+ * same-device VRAM
+ * @cur: The cursor.
+ *
+ * Return: true iff the address returned by xe_res_dma() points to internal vram.
+ */
+static inline bool xe_res_is_vram(const struct xe_res_cursor *cur)
+{
+	if (cur->dma_addr)
+		return cur->dma_addr->proto == XE_INTERCONNECT_VRAM;
+
+	switch (cur->mem_type) {
+	case XE_PL_STOLEN:
+	case XE_PL_VRAM0:
+	case XE_PL_VRAM1:
+		return true;
+	default:
+		break;
+	}
+
+	return false;
 }
 #endif
@@ -34,17 +35,38 @@
 #include "xe_ttm_vram_mgr.h"
 
-/* state back for walking over vram_mgr, stolen_mgr, and gtt_mgr allocations */
+/**
+ * struct xe_res_cursor - state for walking over dma mapping, vram_mgr,
+ * stolen_mgr, and gtt_mgr allocations
+ */
 struct xe_res_cursor {
-- 
2.34.1


  parent reply	other threads:[~2025-03-06  1:26 UTC|newest]

Thread overview: 52+ messages / expand[flat|nested]  mbox.gz  Atom feed  top
2025-03-06  1:26 [PATCH v7 00/32] Introduce GPU SVM and Xe SVM implementation Matthew Brost
2025-03-06  1:26 ` [PATCH v7 01/32] drm/xe: Retry BO allocation Matthew Brost
2025-03-06  1:26 ` [PATCH v7 02/32] mm/migrate: Add migrate_device_pfns Matthew Brost
2025-03-06  1:26 ` [PATCH v7 03/32] mm/migrate: Trylock device page in do_swap_page Matthew Brost
2025-03-06  1:26 ` [PATCH v7 04/32] drm/pagemap: Add DRM pagemap Matthew Brost
2025-03-06  1:26 ` [PATCH v7 05/32] drm/xe/bo: Introduce xe_bo_put_async Matthew Brost
2025-03-06  1:26 ` [PATCH v7 06/32] drm/gpusvm: Add support for GPU Shared Virtual Memory Matthew Brost
2025-03-06  1:26 ` [PATCH v7 07/32] drm/xe: Select DRM_GPUSVM Kconfig Matthew Brost
2025-03-06  1:26 ` [PATCH v7 08/32] drm/xe/uapi: Add DRM_XE_VM_BIND_FLAG_CPU_ADDR_MIRROR Matthew Brost
2025-03-06  1:26 ` [PATCH v7 09/32] drm/xe: Add SVM init / close / fini to faulting VMs Matthew Brost
2025-03-06  1:26 ` Matthew Brost [this message]
2025-03-06  1:26 ` [PATCH v7 11/32] drm/xe: Nuke VM's mapping upon close Matthew Brost
2025-03-06  1:26 ` [PATCH v7 12/32] drm/xe: Add SVM range invalidation and page fault Matthew Brost
2025-03-06  1:26 ` [PATCH v7 13/32] drm/gpuvm: Add DRM_GPUVA_OP_DRIVER Matthew Brost
2025-03-06  1:26 ` [PATCH v7 14/32] drm/xe: Add (re)bind to SVM page fault handler Matthew Brost
2025-03-06  1:26 ` [PATCH v7 15/32] drm/xe: Add SVM garbage collector Matthew Brost
2025-03-06  1:26 ` [PATCH v7 16/32] drm/xe: Add unbind to " Matthew Brost
2025-03-06  1:26 ` [PATCH v7 17/32] drm/xe: Do not allow CPU address mirror VMA unbind if Matthew Brost
2025-03-06  1:26 ` [PATCH v7 18/32] drm/xe: Enable CPU address mirror uAPI Matthew Brost
2025-03-06  1:26 ` [PATCH v7 19/32] drm/xe/uapi: Add DRM_XE_QUERY_CONFIG_FLAG_HAS_CPU_ADDR_MIRROR Matthew Brost
2025-03-06  1:26 ` [PATCH v7 20/32] drm/xe: Add migrate layer functions for SVM support Matthew Brost
2025-03-06  1:26 ` [PATCH v7 21/32] drm/xe: Add SVM device memory mirroring Matthew Brost
2025-03-06  1:26 ` [PATCH v7 22/32] drm/xe: Add drm_gpusvm_devmem to xe_bo Matthew Brost
2025-03-06  1:26 ` [PATCH v7 23/32] drm/xe: Add drm_pagemap ops to SVM Matthew Brost
2025-03-06  1:26 ` [PATCH v7 24/32] drm/xe: Add GPUSVM device memory copy vfunc functions Matthew Brost
2025-03-06  1:26 ` [PATCH v7 25/32] drm/xe: Add Xe SVM populate_devmem_pfn GPU SVM vfunc Matthew Brost
2025-03-06  1:26 ` [PATCH v7 26/32] drm/xe: Add Xe SVM devmem_release " Matthew Brost
2025-03-06  1:26 ` [PATCH v7 27/32] drm/xe: Add SVM VRAM migration Matthew Brost
2025-03-06  1:26 ` [PATCH v7 28/32] drm/xe: Basic SVM BO eviction Matthew Brost
2025-03-06  1:26 ` [PATCH v7 29/32] drm/xe: Add SVM debug Matthew Brost
2025-03-06  1:26 ` [PATCH v7 30/32] drm/xe: Add modparam for SVM notifier size Matthew Brost
2025-03-06  1:26 ` [PATCH v7 31/32] drm/xe: Add always_migrate_to_vram modparam Matthew Brost
2025-03-06  1:26 ` [PATCH v7 32/32] drm/doc: gpusvm: Add GPU SVM documentation Matthew Brost
2025-03-06  5:45   ` Alistair Popple
2025-03-06  6:08     ` Matthew Brost
2025-03-06  1:54 ` ✓ CI.Patch_applied: success for Introduce GPU SVM and Xe SVM implementation (rev7) Patchwork
2025-03-06  1:54 ` ✗ CI.checkpatch: warning " Patchwork
2025-03-06  1:56 ` ✓ CI.KUnit: success " Patchwork
2025-03-06  2:12 ` ✓ CI.Build: " Patchwork
2025-03-06  2:14 ` ✓ CI.Hooks: " Patchwork
2025-03-06  2:16 ` ✗ CI.checksparse: warning " Patchwork
2025-03-06  2:51 ` ✓ Xe.CI.BAT: success " Patchwork
2025-03-06  8:25 ` ✗ Xe.CI.Full: failure " Patchwork
2025-03-06  9:57   ` Matthew Brost
2025-03-06  9:22 ` ✓ CI.Patch_applied: success for Introduce GPU SVM and Xe SVM implementation (rev8) Patchwork
2025-03-06  9:23 ` ✗ CI.checkpatch: warning " Patchwork
2025-03-06  9:24 ` ✓ CI.KUnit: success " Patchwork
2025-03-06  9:41 ` ✓ CI.Build: " Patchwork
2025-03-06  9:43 ` ✓ CI.Hooks: " Patchwork
2025-03-06  9:44 ` ✗ CI.checksparse: warning " Patchwork
2025-03-06 10:17 ` ✗ Xe.CI.BAT: failure " Patchwork
2025-03-06 18:41 ` ✗ Xe.CI.Full: " Patchwork

Reply instructions:

You may reply publicly to this message via plain-text email
using any one of the following methods:

* Save the following mbox file, import it into your mail client,
  and reply-to-all from there: mbox

  Avoid top-posting and favor interleaved quoting:
  https://en.wikipedia.org/wiki/Posting_style#Interleaved_style

* Reply using the --to, --cc, and --in-reply-to
  switches of git-send-email(1):

  git send-email \
    --in-reply-to=20250306012657.3505757-11-matthew.brost@intel.com \
    --to=matthew.brost@intel.com \
    --cc=airlied@gmail.com \
    --cc=apopple@nvidia.com \
    --cc=dakr@kernel.org \
    --cc=dri-devel@lists.freedesktop.org \
    --cc=felix.kuehling@amd.com \
    --cc=himal.prasad.ghimiray@intel.com \
    --cc=intel-xe@lists.freedesktop.org \
    --cc=simona.vetter@ffwll.ch \
    --cc=thomas.hellstrom@linux.intel.com \
    /path/to/YOUR_REPLY

  https://kernel.org/pub/software/scm/git/docs/git-send-email.html

* If your mail client supports setting the In-Reply-To header
  via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox